@@ -51,11 +51,11 @@ static struct gvm_configuration config;
 static LibHalContext *hal_ctx;
 
 #ifndef ASSUME_SUBMOUNT
-/** List of UDI's for volumes mounted by g-v-m that we need to apply policy to*/
-static GSList *mounted_volumes_policy_queue = NULL;
+/** Table of UDI's for volumes being mounted by g-v-m that we need to apply policy to */
+static GHashTable *mount_table = NULL;
 
 /** List of UDI's of all volumes mounted during the lifetime of the program */
-static GSList *all_mounted_volumes = NULL;
+static GSList *mounted_volumes = NULL;
 #endif /* ! ASSUME_SUBMOUNT */
 
 
@@ -856,15 +856,18 @@ gvm_device_mounted (const char *udi)
 			gvm_run_camera (udi, device, mount_point);
 	} else if (config.autoipod && gvm_udi_is_ipod (udi)) {
 		int action = GVM_RESPONSE_SYNC_MUSIC;
+		char *ipod_control;
 		
-		if (gvm_check_photos (udi, device, mount_point)) {
+		ipod_control = g_build_filename (mount_point, "iPod_Control", NULL);
+		
+		if (gvm_check_photos (udi, device, ipod_control)) {
 			/* we have ourselves an iPod Photo - need to prompt what to do */
 			action = gvm_prompt (GVM_PROMPT_IPOD_PHOTO);
 		}
 		
 		switch (action) {
 		case GVM_RESPONSE_IMPORT_PHOTOS:
-			gvm_run_camera (udi, device, mount_point);
+			gvm_run_camera (udi, device, ipod_control);
 			break;
 		case GVM_RESPONSE_SYNC_MUSIC:
 			gvm_run_ipod (udi, device, mount_point);
@@ -872,6 +875,8 @@ gvm_device_mounted (const char *udi)
 		default:
 			break;
 		}
+		
+		g_free (ipod_control);
 	} else {
 		gvm_autorun (udi, device, mount_point);
 	}
@@ -925,7 +930,7 @@ gvm_run_cdplayer (const char *udi, const char *device, const char *mount_point)
 static void
 gvm_ask_mixed (const char *udi)
 {
-	char *device = NULL, *mount_point = NULL;
+	char *key, *device = NULL, *mount_point = NULL;
 	int action = GVM_RESPONSE_NONE;
 	DBusError error;
 	
@@ -951,7 +956,9 @@ gvm_ask_mixed (const char *udi)
 	case GVM_RESPONSE_BROWSE:
 #ifndef ASSUME_SUBMOUNT
 		gvm_device_mount (udi, device, NULL);
-		mounted_volumes_policy_queue = g_slist_append (mounted_volumes_policy_queue, g_strdup (udi));
+		
+		key = g_strdup (udi);
+		g_hash_table_insert (mount_table, key, key);
 #else
 		gvm_device_mounted (udi);
 #endif
@@ -1059,8 +1066,12 @@ gvm_cdrom_policy (const char *udi)
 	} else if (has_data) {
 #ifndef ASSUME_SUBMOUNT
 		if (config.automount_media) {
+			char *key;
+			
 			gvm_device_mount (udi, device, NULL);
-			mounted_volumes_policy_queue = g_slist_append (mounted_volumes_policy_queue, g_strdup (udi));
+			
+			key = g_strdup (udi);
+			g_hash_table_insert (mount_table, key, key);
 		}
 #else
 		gvm_device_mounted (udi);
@@ -1086,51 +1097,76 @@ gvm_cdrom_policy (const char *udi)
  * drive, we enforce the appropriate policy.
  *
  * At the moment, we only handle CD-ROM and DVD drives.
+ *
+ * Returns TRUE if the device was handled or FALSE otherwise
  */
-static void
-gvm_media_changed (const char *udi, const char *storage_device, const char *device)
+static gboolean
+gvm_media_changed (const char *udi, const char *storage_device)
 {
+	gboolean handled = FALSE;
 	char *media_type;
 	DBusError error;
-
+	
 	/* Refuse to enforce policy on removable media if drive is locked */
 	dbus_error_init (&error);
 	if (libhal_device_property_exists (hal_ctx, storage_device, "info.locked", NULL)
 	    && libhal_device_get_property_bool (hal_ctx, storage_device, "info.locked", NULL)) {
 		dbg ("Drive with udi %s is locked through hal; skipping policy\n", storage_device);
-		return;
+		/* we return TRUE here because the device is locked - we can pretend we handled it */
+		return TRUE;
 	}
 	
 	/*
 	 * Get HAL's interpretation of our media type.  Note that we must check
 	 * the storage device and not this UDI
 	 */
-	media_type = libhal_device_get_property_string (hal_ctx, storage_device,
-						        "storage.drive_type",
-						        &error);
-	if (!media_type) {
+	if (!(media_type = libhal_device_get_property_string (hal_ctx, storage_device, "storage.drive_type", &error))) {
 		warn ("cannot get storage.drive_type: %s", error.message);
 		if (dbus_error_is_set (&error))
 			dbus_error_free (&error);
-		return;
+		
+		/* FIXME: should we really return TRUE here? not sure... need a test case */
+		
+		return TRUE;
 	}
 	
-	if (!g_ascii_strcasecmp (media_type, "cdrom")) {
+	if (!strcmp (media_type, "cdrom")) {
 		gvm_cdrom_policy (udi);
-	} else {
-#ifndef ASSUME_SUBMOUNT
-		dbg ("Auto-mounting: %s\n", device);
-		
-		if (config.automount_drives) {
-			gvm_device_mount (udi, device, NULL);
-			mounted_volumes_policy_queue = g_slist_append (mounted_volumes_policy_queue, g_strdup (udi));
-		}
-#else
-		gvm_device_mounted (udi);
-#endif
+		handled = TRUE;
 	}
 	
 	libhal_free_string (media_type);
+	
+	return handled;
+}
+
+
+/* Return whether or not the device should be automounted based on the
+ * HAL automount hint. Default to returning TRUE if the property isn't
+ * set for the device or its parent storage device. */
+static gboolean
+gvm_automount_enabled (const char *udi)
+{
+	gboolean enabled = TRUE;
+	DBusError error;
+	char *drive;
+	
+	if (libhal_device_property_exists (hal_ctx, udi, "storage.automount_enabled_hint", NULL))
+		return libhal_device_get_property_bool (hal_ctx, udi, "storage.automount_enabled_hint", NULL);
+	
+	dbus_error_init (&error);
+	if (!(drive = libhal_device_get_property_string (hal_ctx, udi, "info.parent", &error))) {
+		warn ("couldn't get info.parent for %s: %s", udi, error.message);
+		dbus_error_free (&error);
+		return TRUE; /* sensible default? */
+	}
+	
+	if (libhal_device_property_exists (hal_ctx, drive, "storage.automount_enabled_hint", NULL))
+		enabled = libhal_device_get_property_bool (hal_ctx, drive, "storage.automount_enabled_hint", NULL);
+	
+	libhal_free_string (drive);
+	
+	return enabled;
 }
 
 /** Invoked when a device is added to the Global Device List. 
@@ -1180,17 +1216,25 @@ hal_device_added (LibHalContext *ctx __attribute__((__unused__)),
 		 * Does this device support removable media?  Note that we
 		 * check storage_device and not our own UDI
 		 */
-		if (mountable && libhal_device_get_property_bool (hal_ctx, storage_device, "storage.removable", NULL)) {
+		if (libhal_device_get_property_bool (hal_ctx, storage_device, "storage.removable", NULL)) {
 			/* we handle media change events separately */
 			dbg ("Changed: %s\n", device);
-			gvm_media_changed (udi, storage_device, device);
-			goto out;
+			if (gvm_media_changed (udi, storage_device))
+				goto out;
 		}
 		
 #ifndef ASSUME_SUBMOUNT
 		if (config.automount_drives && mountable) {
-			gvm_device_mount (udi, device, NULL);
-			mounted_volumes_policy_queue = g_slist_append (mounted_volumes_policy_queue, g_strdup (udi));
+			if (gvm_automount_enabled (udi)) {
+				char *key;
+				
+				gvm_device_mount (udi, device, NULL);
+				
+				key = g_strdup (udi);
+				g_hash_table_insert (mount_table, key, key);
+			} else {
+				dbg ("storage.automount_enabled_hint set to false on %s, not mounting\n", udi);
+			}
 		}
 #else
 		if (mountable)
@@ -1269,46 +1313,37 @@ hal_property_modified (LibHalContext *ctx __attribute__((__unused__)),
 		       dbus_bool_t is_added __attribute__((__unused__)))
 {
 #ifndef ASSUME_SUBMOUNT
+	char *mounting_udi;
 	dbus_bool_t val;
-	GSList *i;
-	GSList *next;
-
-	if (g_ascii_strcasecmp (key, "volume.is_mounted") != 0)
+	GSList *l, *n;
+	
+	if (strcmp (key, "volume.is_mounted") != 0)
 		return;
 	
 	val = libhal_device_get_property_bool (hal_ctx, udi, key, NULL);
 	
 	if (val == TRUE) {
-		GSList *policy_udi;
-
 		dbg ("Mounted: %s\n", udi);
-
+		
 		/* add to list of all volumes mounted during lifetime */
-		all_mounted_volumes = g_slist_append (all_mounted_volumes,
-						      g_strdup (udi));
-
-		policy_udi = g_slist_find_custom (mounted_volumes_policy_queue, udi, 
-						  (GCompareFunc) g_ascii_strcasecmp);
-		if (policy_udi != NULL) {
-			g_free (policy_udi->data);
-			mounted_volumes_policy_queue = g_slist_delete_link (mounted_volumes_policy_queue, 
-									    policy_udi);
+		mounted_volumes = g_slist_append (mounted_volumes, g_strdup (udi));
+		
+		if ((mounting_udi = g_hash_table_lookup (mount_table, udi))) {
+			g_hash_table_remove (mount_table, udi);
+			g_free (mounting_udi);
 			
 			gvm_device_mounted (udi);
 		}
-
 	} else {
 		dbg ("Unmounted: %s\n", udi);
-
+		
 		/* remove from list of all volumes mounted during lifetime */
-
-		for (i=all_mounted_volumes; i != NULL; i = next) {
-			next = g_slist_next (i);
-			if (strcmp (udi, (const char *)i->data) == 0) {
-				g_free (i->data);
-				all_mounted_volumes = 
-					g_slist_delete_link (
-						all_mounted_volumes, i);
+		
+		for (l = mounted_volumes; l != NULL; l = n) {
+			n = g_slist_next (l);
+			if (strcmp (udi, (const char *) l->data) == 0) {
+				g_free (l->data);
+				mounted_volumes = g_slist_delete_link (mounted_volumes, l);
 				break;
 			}
 		}
@@ -1430,8 +1465,8 @@ gvm_do_hal_init (void)
 static void
 mount_all (LibHalContext *ctx)
 {
-	char *prop, *dev, *udi;
-	int num_volumes;
+	char *prop, *dev, *udi, *drive;
+	int num_volumes, mount;
 	char **volumes;
 	DBusError error;
 	int i;
@@ -1465,6 +1500,23 @@ mount_all (LibHalContext *ctx)
 		}
 		libhal_free_string (prop);
 		
+		/* check our mounting policy */
+		if (!(drive = libhal_device_get_property_string (ctx, udi, "info.parent", NULL)))
+			continue;
+		
+		if (libhal_device_property_exists (ctx, drive, "storage.hotpluggable", NULL)
+		    && libhal_device_get_property_bool (ctx, drive, "storage.hotpluggable", NULL))
+			mount = config.automount_drives;
+		else if (libhal_device_property_exists (ctx, drive, "storage.removable", NULL)
+			 && libhal_device_get_property_bool (ctx, drive, "storage.removable", NULL))
+			mount = config.automount_media;
+		else
+			mount = TRUE;
+		libhal_free_string (drive);
+		
+		if (!mount)
+			continue;
+		
 		/* mount the device */
 		if (!(dev = libhal_device_get_property_string (ctx, udi, "block.device", &error))) {
 			dbg ("mount_all: mounting %s\n", dev);
@@ -1494,7 +1546,7 @@ unmount_all (LibHalContext *ctx)
 	
 	dbg ("unmounting all volumes that we saw mounted in our life\n");
 	dbus_error_init (&error);
-	for (l = all_mounted_volumes; l != NULL; l = g_slist_next (l)) {
+	for (l = mounted_volumes; l != NULL; l = g_slist_next (l)) {
 		udi = l->data;
 		
 		device = libhal_device_get_property_string (ctx, udi, "block.device", &error);
@@ -1574,14 +1626,14 @@ int
 main (int argc, char *argv[])
 {
 	GnomeClient *client;
-
+	
 	gnome_program_init (PACKAGE, VERSION, LIBGNOMEUI_MODULE,
 			    argc, argv, GNOME_PARAM_NONE);
-
-	bindtextdomain(PACKAGE, GNOMELOCALEDIR);
-	bind_textdomain_codeset(PACKAGE, "UTF-8");
-	textdomain(PACKAGE);
-
+	
+	bindtextdomain (PACKAGE, GNOMELOCALEDIR);
+	bind_textdomain_codeset (PACKAGE, "UTF-8");
+	textdomain (PACKAGE);
+	
 	client = gnome_master_client ();
 	if (gvm_get_clipboard ()) {
 		gnome_client_set_restart_style (client, GNOME_RESTART_ANYWAY);
@@ -1596,7 +1648,7 @@ main (int argc, char *argv[])
 	hal_ctx = gvm_do_hal_init ();
 	if (!hal_ctx)
 		return 1;
-
+	
 	gvm_init_config ();
 	
 	/* SIGTERM handling via pipes  */
@@ -1604,19 +1656,23 @@ main (int argc, char *argv[])
 		warn ("Could not setup pipe, errno=%d", errno);
 		return 1;
 	}
+	
 	sigterm_iochn = g_io_channel_unix_new (sigterm_unix_signal_pipe_fds[0]);
 	if (sigterm_iochn == NULL) {
 		warn ("Could not create GIOChannel");
 		return 1;
 	}
+	
 	g_io_add_watch (sigterm_iochn, G_IO_IN, sigterm_iochn_data, NULL);
 	signal (SIGTERM, handle_sigterm);
 	
+	mount_table = g_hash_table_new (g_str_hash, g_str_equal);
+	
 #ifndef ASSUME_SUBMOUNT
 	mount_all (hal_ctx);
 #endif
 	
 	gtk_main ();
-
+	
 	return 0;
 }
