@@ -443,7 +443,9 @@ main (int argc, char **argv)
 		gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
 							  _("The \"hald\" service is required but not currently "
 							    "running. Enable the service and rerun this applet, "
-							    "or contact your system administrator."));
+							    "or contact your system administrator.\n\n"
+							    "Note: You need Linux kernel 2.6 for volume "
+							    "management to work."));
 		
 		gtk_dialog_run (GTK_DIALOG (dialog));
 		gtk_widget_destroy (dialog);
